/*
 * Copyright © 2021 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of ipspiel21-tictactoe-core.
 *
 * ipspiel21-tictactoe-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * ipspiel21-tictactoe-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * ipspiel21-tictactoe-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.ipspiel21.viergewinnt.core.domain;


import java.util.List;
import java.util.Map;
import java.util.Set;

import de.fhdw.gaming.core.domain.Stateful;

/**
 * Represents the VierGewinnt board.
 * <p>
 * A VierGewinnt board is a square and hence possesses an identical non-zero
 * positive number of rows and columns. Each combination of a row and a column
 * is described by a {@link VierGewinntPosition position}. At each position,
 * there is a {@link VierGewinntField field} which describes whether the field
 * is empty or marked by a cross or nought.
 */
public interface VierGewinntBoard extends Stateful {

    /**
     * Returns the number of fields of this board.
     */
    int getSize();

    /**
     * Returns the number of rows of this board.
     */
    int getRowSize();

    /**
     * Returns the number of columns of this board.
     */
    int getColumnSize();

    /**
     * Returns a list of all rows on this board.
     */
    List<VierGewinntRow> getRows();

    /**
     * Checks whether this board contains a field at the position passed.
     *
     * @param position The position.
     * @return {@code true} if this board contains a field at the position passed,
     *         else {@code false}.
     */
    boolean hasFieldAt(VierGewinntPosition position);

    /**
     * Returns the field at the given position.
     *
     * @param position The position of the field to return.
     * @return The field.
     * @throws IllegalArgumentException if the position is out of range, i.e. if it
     *                                  does not denote a field.
     */
    VierGewinntField getFieldAt(VierGewinntPosition position);

    /**
     * Returns all fields of this board line by line.
     * <p>
     * The set returned are possibly immutable, so do not modify them.
     */
    List<VierGewinntField> getFields();

    /**
     * Returns all fields of a given state.
     * <p>
     * The map returned returned is possibly immutable, so do not modify it.
     *
     * @param fieldState The state of the fields to return.
     * @return The fields.
     */
    Map<VierGewinntPosition, ? extends VierGewinntField> getFieldsBeing(VierGewinntFieldState fieldState);

    /**
     * All playable fields.
     * 
     * @return a Set of {@link VierGewinntField} 
     */
    Set<VierGewinntField> getAllPlayableFields();

    @Override
    VierGewinntBoard deepCopy();
    
}
