/*
 * Copyright © 2021 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of ipspiel21-tictactoe-core.
 *
 * ipspiel21-tictactoe-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * ipspiel21-tictactoe-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * ipspiel21-tictactoe-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.ipspiel21.viergewinnt.core.domain;

import de.fhdw.gaming.core.domain.GameException;

/**
 * Represents a single field on a {@link VierGewinntBoard board}.
 * <p>
 * A field has a {@link VierGewinntPosition position}. This position is fixed and does not change.
 * <p>
 * A field has a {@link VierGewinntFieldState state}. The state may change, but only the following state transitions are
 * allowed:
 * <ul>
 * <li>{@link VierGewinntFieldState#EMPTY EMPTY} -&gt; {@link VierGewinntFieldState#CROSS CROSS}</li>
 * <li>{@link VierGewinntFieldState#EMPTY EMPTY} -&gt; {@link VierGewinntFieldState#NOUGHT NOUGHT}</li>
 * </ul>
 * In other words, a non-empty field may never change again.
 * <p>
 * The state of a field cannot be changed directly. Instead, a conforming move has to be submitted to the rule engine.
 */
public interface VierGewinntField {

    /**
     * Returns the board this field belongs to.
     */
    VierGewinntBoard getBoard();

    /**
     * Returns the position of this field.
     */
    VierGewinntPosition getPosition();

    /**
     * Returns the current state of this field.
     */
    VierGewinntFieldState getState();

    /**
     * Determines whether there is a neighbour field in a given direction.
     *
     * @param direction The direction to use.
     * @return {@code true} if there is a neighbour field in the given direction, otherwise {@code false}.
     */
    boolean hasNeighbour(VierGewinntDirection direction);

    /**
     * Returns a neighbour field in a given direction.
     *
     * @param direction The direction to use.
     * @return The corresponding neighbour field.
     * @throws IllegalArgumentException if there is no neighbour field in the given direction.
     */
    VierGewinntField getNeighbour(VierGewinntDirection direction) throws IllegalArgumentException;

    /**
     * Changes the state of this field. Requires this field to be empty.
     *
     * @param state The new state of the field.
     * @throws GameException if placing a token of the given type is not allowed according to the rules of the game.
     */
    void changeState(VierGewinntFieldState state) throws GameException;

    /**
     * Returns an equal copy of this object.
     *
     * @param target The board to copy this field to.
     */
    VierGewinntField deepCopy(VierGewinntBoard target);

}
