package de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import de.fhdw.gaming.core.domain.GameException;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntBoard;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntField;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntFieldState;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntPosition;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntRow;

/**
 * Implements {@link VierGewinntBoard}.
 *
 * @author Leonard Frank
 */
public class VierGewinntBoardImpl implements VierGewinntBoard {

    /**
     * Contains all rows of a {@link VierGewinntBoard}.
     */
    private final List<VierGewinntRow> rows;

    /**
     * Number of rows on board.
     */
    private final int rowSize;

    /**
     * Number of columns on board.
     */
    private final int columnSize;

    /**
     * Creates a VierGewinntBoard.
     *
     * @param rowSize    Number of rows on this board
     * @param columnSize Defines the length of each row on this board
     * @throws GameException
     */
    public VierGewinntBoardImpl(final int rowSize, final int columnSize)
            throws IllegalArgumentException, GameException {
        if (rowSize < 4 || columnSize < 4) {
            throw new IllegalArgumentException("Row or column size is smaller than 4");
        }

        this.columnSize = columnSize;
        this.rowSize = rowSize;
        this.rows = new ArrayList<>();
        for (int rowIndex = 0; rowIndex < rowSize; rowIndex++) {
            this.rows.add(new ViergewinntRowImpl(columnSize, rowIndex, this));
        }
    }

    /**
     * Creates a board from an existing board.
     *
     * @param board
     */
    public VierGewinntBoardImpl(final VierGewinntBoard board) {
        this.columnSize = board.getColumnSize();
        this.rowSize = board.getRowSize();
        this.rows = new ArrayList<>();
        for (final VierGewinntRow row : board.getRows()) {
            this.rows.add(row.deepCopy(this));
        }
    }

    @Override
    public int getSize() {
        return this.rowSize * this.columnSize;
    }

    @Override
    public int getRowSize() {
        return this.rowSize;
    }

    @Override
    public int getColumnSize() {
        return this.columnSize;
    }

    @Override
    public List<VierGewinntRow> getRows() {
        return this.rows;
    }

    @Override
    public boolean hasFieldAt(final VierGewinntPosition position) {
        for (final VierGewinntRow row : this.rows) {
            for (final VierGewinntField field : row.getFields()) {
                if (field.getPosition().equals(position)) {
                    return true;
                }
            }
        }
        return false;
    }

    @Override
    public VierGewinntField getFieldAt(final VierGewinntPosition position) throws IllegalArgumentException {
        try {
            this.rows.get(position.getRow());
        } catch (final IndexOutOfBoundsException e) {
            throw new IllegalArgumentException("There is no row for this position", e);
        }
        final VierGewinntRow requestedRow = this.rows.get(position.getRow());

        for (final VierGewinntField field : requestedRow.getFields()) {
            if (field.getPosition().equals(position)) {
                return field;
            }
        }
        throw new IllegalArgumentException("There is no field with this position in the row");
    }

    @Override
    public List<VierGewinntField> getFields() {
        final List<VierGewinntField> allFields = new ArrayList<>();
        for (final VierGewinntRow row : this.rows) {
            for (final VierGewinntField field : row.getFields()) {
                allFields.add(field);
            }

        }
        return allFields;
    }

    @Override
    public Map<VierGewinntPosition, ? extends VierGewinntField>
            getFieldsBeing(final VierGewinntFieldState fieldState) {
        final HashMap<VierGewinntPosition, VierGewinntField> result = new HashMap<>();
        for (final VierGewinntRow row : this.rows) {
            for (final VierGewinntField field : row.getFields()) {
                if (field.getState().equals(fieldState)) {
                    result.put(field.getPosition(), field);
                }
            }
        }
        return result;
    }

    @Override
    public Set<VierGewinntField> getAllPlayableFields() {
        final Map<Integer, VierGewinntField> result = new LinkedHashMap<>();

        final ListIterator<VierGewinntRow> rowIt = this.rows.listIterator(this.rows.size());
        while (rowIt.hasPrevious()) {
            final VierGewinntRow row = rowIt.previous();
            final Iterator<? extends VierGewinntField> fieldIt = row.getFields().iterator();
            for (int column = 0; column < row.getFields().size(); ++column) {
                final VierGewinntField field = fieldIt.next();
                if (result.containsKey(column)) {
                    continue;
                }
                if (field.getState().equals(VierGewinntFieldState.EMPTY)) {
                    result.put(column, field);
                }
            }
        }
        return new LinkedHashSet<>(result.values());
    }

    @Override
    public VierGewinntBoard deepCopy() {
        return new VierGewinntBoardImpl(this);
    }

    @Override
    public String toString() {
        final StringBuilder out = new StringBuilder("  ");
        for (final VierGewinntField field : this.rows.get(0).getFields()) {
            out.append(field.getPosition().toString().substring(0, 1)).append(' ');
        }
        out.append('\n');
        for (final VierGewinntRow row : this.rows) {
            out.append(row.getFields().get(0).getPosition().toString().substring(1, 2)).append(' ')
                    .append(row.toString()).append('\n');
        }
        return out.toString();
    }

    @Override
    public boolean equals(final Object obj) {
        if (obj == null || obj.getClass() != this.getClass()) {
            return false;
        }
        final VierGewinntBoardImpl otherBoard = (VierGewinntBoardImpl) obj;
        if (otherBoard.getRowSize() != this.getRowSize() || otherBoard.getColumnSize() != this.getColumnSize()) {
            return false;
        }
        final Iterator<VierGewinntRow> otherRows = otherBoard.getRows().iterator();
        final Iterator<VierGewinntRow> thisRows = this.getRows().iterator();
        while (otherRows.hasNext() && thisRows.hasNext()) {
            if (!(otherRows.next().equals(thisRows.next()))) {
                return false;
            }
        }
        return true;
    }

    /**
     * Returns HashCode of this.
     */
    @Override
    public int hashCode() {
        return Objects.hash(this.rows, this.rowSize, this.columnSize);
    }

}
