/*
 * Copyright © 2021 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of ipspiel21-viergewinnt-core.
 *
 * ipspiel21-viergewinnt-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * ipspiel21-viergewinnt-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * ipspiel21-viergewinnt-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.impl;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.regex.Pattern;

import de.fhdw.gaming.core.domain.GameBuilder;
import de.fhdw.gaming.core.domain.GameBuilderFactory;
import de.fhdw.gaming.core.domain.GameException;
import de.fhdw.gaming.core.domain.Strategy;
import de.fhdw.gaming.core.ui.InputProvider;
import de.fhdw.gaming.core.ui.InputProviderException;
import de.fhdw.gaming.core.ui.type.validator.MaxValueValidator;
import de.fhdw.gaming.core.ui.type.validator.MinValueValidator;
import de.fhdw.gaming.core.ui.type.validator.PatternValidator;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntGameBuilder;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntGameBuilderFactory;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntPlayerBuilder;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntStrategy;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.factory.VierGewinntDefaultStrategyFactoryProvider;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.factory.VierGewinntStrategyFactory;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.factory.VierGewinntStrategyFactoryProvider;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.moves.factory.VierGewinntDefaultMoveFactory;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.moves.factory.VierGewinntMoveFactory;

/**
 * Implements {@link VierGewinntGameBuilderFactory} by creating a Vier Gewinnt game builder.
 */
public final class VierGewinntGameBuilderFactoryImpl implements VierGewinntGameBuilderFactory {

    /**
     * The number of players.
     */
    private static final int NUMBER_OF_PLAYERS = 2;
    /**
     * Minimum number of rows of the board.
     */
    private static final int MIN_ROW_SIZE = 4;
    /**
     * Minimum number of columns of the board.
     */
    private static final int MIN_COLUMN_SIZE = 4;
    /**
     * Maximum number of rows of the board.
     */
    private static final int MAX_ROW_SIZE = 7;
    /**
     * Maximum number of rows of the board.
     */
    private static final int MAX_COLUMN_SIZE = 7;
    /**
     * Smallest allowed maximum computation time per move in seconds.
     */
    private static final int MIN_MAX_COMPUTATION_TIME_PER_MOVE = 1;
    /**
     * Largest allowed maximum computation time per move in seconds.
     */
    private static final int MAX_MAX_COMPUTATION_TIME_PER_MOVE = 3600;

    /**
     * All available Vier Gewinnt strategies.
     */
    private final Set<VierGewinntStrategy> strategies;

    /**
     * Creates an Vier Gewinnt game factory. Vier Gewinnt strategies are loaded by using the
     * {@link java.util.ServiceLoader}.
     * <p>
     * This constructor is meant to be used by the {@link java.util.ServiceLoader}.
     */
    public VierGewinntGameBuilderFactoryImpl() {
        this(new VierGewinntDefaultStrategyFactoryProvider());
    }

    /**
     * Creates an Vier Gewinnt game factory.
     *
     * @param strategyFactoryProvider The {@link VierGewinntStrategyFactoryProvider} for loading Vier Gewinnt
     *                                strategies.
     */
    VierGewinntGameBuilderFactoryImpl(final VierGewinntStrategyFactoryProvider strategyFactoryProvider) {
        final VierGewinntMoveFactory moveFactory = new VierGewinntDefaultMoveFactory();

        final List<VierGewinntStrategyFactory> factories = strategyFactoryProvider.getStrategyFactories();
        this.strategies = new LinkedHashSet<>();
        for (final VierGewinntStrategyFactory factory : factories) {
            this.strategies.add(factory.create(moveFactory));
        }
    }

    @Override
    public String getName() {
        return "VierGewinnt";
    }

    @Override
    public int getMinimumNumberOfPlayers() {
        return VierGewinntGameBuilderFactoryImpl.NUMBER_OF_PLAYERS;
    }

    @Override
    public int getMaximumNumberOfPlayers() {
        return VierGewinntGameBuilderFactoryImpl.NUMBER_OF_PLAYERS;
    }

    @Override
    public List<? extends Strategy<?, ?, ?>> getStrategies() {
        return new ArrayList<>(this.strategies);
    }

    @Override
    public VierGewinntGameBuilder createGameBuilder(final InputProvider inputProvider) throws GameException {
        try {
            final VierGewinntGameBuilder gameBuilder = new VierGewinntGameBuilderImpl();

            @SuppressWarnings("unchecked")
            final Map<String, Object> gameData = inputProvider
                    .needInteger(VierGewinntGameBuilderFactory.PARAM_COLUMN_SIZE, "Number of columns",
                            Optional.of(VierGewinntGameBuilder.DEFAULT_COLUMN_SIZE),
                            new MinValueValidator<>(VierGewinntGameBuilderFactoryImpl.MIN_COLUMN_SIZE),
                            new MaxValueValidator<>(VierGewinntGameBuilderFactoryImpl.MAX_COLUMN_SIZE))
                    .needInteger(VierGewinntGameBuilderFactory.PARAM_ROW_SIZE, "Number of rows",
                            Optional.of(VierGewinntGameBuilder.DEFAULT_ROW_SIZE),
                            new MinValueValidator<>(VierGewinntGameBuilderFactoryImpl.MIN_ROW_SIZE),
                            new MaxValueValidator<>(VierGewinntGameBuilderFactoryImpl.MAX_ROW_SIZE))
                    .needInteger(GameBuilderFactory.PARAM_MAX_COMPUTATION_TIME_PER_MOVE,
                            "Maximum computation time per move in seconds",
                            Optional.of(GameBuilder.DEFAULT_MAX_COMPUTATION_TIME_PER_MOVE),
                            new MinValueValidator<>(
                                    VierGewinntGameBuilderFactoryImpl.MIN_MAX_COMPUTATION_TIME_PER_MOVE),
                            new MaxValueValidator<>(
                                    VierGewinntGameBuilderFactoryImpl.MAX_MAX_COMPUTATION_TIME_PER_MOVE))
                    .requestData("Board properties");

            gameBuilder.changeColumnSize((Integer) gameData.get(VierGewinntGameBuilderFactory.PARAM_COLUMN_SIZE));
            gameBuilder.changeRowSize((Integer) gameData.get(VierGewinntGameBuilderFactory.PARAM_ROW_SIZE));
            gameBuilder.changeMaximumComputationTimePerMove(
                    (Integer) gameData.get(GameBuilderFactory.PARAM_MAX_COMPUTATION_TIME_PER_MOVE));

            final InputProvider firstPlayerInputProvider = inputProvider.getNext(gameData);
            final Map<String, Object> firstPlayerData = this.requestPlayerData(firstPlayerInputProvider, "Player 1",
                    Optional.empty());
            final VierGewinntPlayerBuilder firstPlayerBuilder = this
                    .initPlayerBuilder(gameBuilder.createPlayerBuilder(), firstPlayerData);
            final VierGewinntStrategy firstPlayerStrategy = this.getStrategy(firstPlayerData);
            gameBuilder.addPlayerBuilder(firstPlayerBuilder, firstPlayerStrategy);

            final InputProvider secondPlayerInputProvider = firstPlayerInputProvider.getNext(firstPlayerData);
            final Map<String, Object> secondPlayerData = this.requestPlayerData(secondPlayerInputProvider, "Player 2",
                    Optional.of(
                            !(Boolean) firstPlayerData.get(VierGewinntGameBuilderFactory.PARAM_PLAYER_USING_YELLOW)));
            final VierGewinntPlayerBuilder secondPlayerBuilder = this
                    .initPlayerBuilder(gameBuilder.createPlayerBuilder(), secondPlayerData);
            final VierGewinntStrategy secondPlayerStrategy = this.getStrategy(secondPlayerData);
            gameBuilder.addPlayerBuilder(secondPlayerBuilder, secondPlayerStrategy);

            return gameBuilder;
        } catch (final InputProviderException e) {
            throw new GameException(String.format("Creating Vier Gewinnt game was aborted: %s", e.getMessage()), e);
        }
    }

    /**
     * Initialises an Vier Gewinnt player builder.
     *
     * @param inputProvider The input provider.
     * @param title         The title for the UI.
     * @param usingYellow   If set, determines whether the player uses yellow or red. Otherwise, the setting is
     *                      initially set to crosses, and the user can change it at will.
     * @return {@code playerBuilder}.
     * @throws InputProviderException if the operation has been aborted prematurely (e.g. if the user cancelled a
     *                                dialog).
     */
    @SuppressWarnings("unchecked")
    private Map<String, Object> requestPlayerData(final InputProvider inputProvider, final String title,
            final Optional<Boolean> usingYellow) throws GameException, InputProviderException {

        inputProvider
                .needString(GameBuilderFactory.PARAM_PLAYER_NAME, "Name", Optional.empty(),
                        new PatternValidator(Pattern.compile("\\S+(\\s+\\S+)*")))
                .needBoolean(VierGewinntGameBuilderFactory.PARAM_PLAYER_USING_YELLOW, "Uses yellow",
                        Optional.of(Boolean.TRUE))
                .needObject(GameBuilderFactory.PARAM_PLAYER_STRATEGY, "Strategy", Optional.empty(), this.strategies);

        if (usingYellow.isPresent()) {
            inputProvider.fixedBoolean(VierGewinntGameBuilderFactory.PARAM_PLAYER_USING_YELLOW, usingYellow.get());
        }

        return inputProvider.requestData(title);
    }

    /**
     * Initialises an Vier Gewinnt player builder.
     *
     * @param playerBuilder The player builder.
     * @param playerData    The requested player data.
     * @return {@code playerBuilder}.
     * @throws InputProviderException if the operation has been aborted prematurely (e.g. if the user cancelled a
     *                                dialog).
     */
    private VierGewinntPlayerBuilder initPlayerBuilder(final VierGewinntPlayerBuilder playerBuilder,
            final Map<String, Object> playerData) throws GameException, InputProviderException {

        return playerBuilder.changeName((String) playerData.get(GameBuilderFactory.PARAM_PLAYER_NAME))
                .changeUsingYellow((Boolean) playerData.get(VierGewinntGameBuilderFactory.PARAM_PLAYER_USING_YELLOW));
    }

    /**
     * Returns an Vier Gewinnt strategy.
     *
     * @param playerData The requested player data.
     * @return The Vier Gewinnt strategy.
     */
    private VierGewinntStrategy getStrategy(final Map<String, Object> playerData) {
        return (VierGewinntStrategy) playerData.get(GameBuilderFactory.PARAM_PLAYER_STRATEGY);
    }
}
