package de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.impl;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

import de.fhdw.gaming.core.domain.DefaultObserverFactoryProvider;
import de.fhdw.gaming.core.domain.GameBuilder;
import de.fhdw.gaming.core.domain.GameException;
import de.fhdw.gaming.core.domain.ObserverFactoryProvider;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntGame;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntGameBuilder;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntPlayerBuilder;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntState;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntStrategy;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.moves.impl.AbstractMove;

/**
 * The Game Builder for a VierGewinnt Game.
 * 
 * @author Robby Rabbitman
 *
 */
public class VierGewinntGameBuilderImpl implements VierGewinntGameBuilder {

    /**
     * Time for a Player to choose his next move.
     * 
     */
    private int maxComputationTimePerMove;
    /**
     * Observer Factory Provider.
     * 
     */
    private ObserverFactoryProvider newObserverFactoryProvider;
    /**
     * The Row Size of a VierGewinnt Game.
     * 
     */
    private int rowSize;
    /**
     * The Column Size of a VierGewinnt Game.
     * 
     */
    private int columnSize;
    /**
     * player builder 1.
     */
    private Optional<VierGewinntPlayerBuilder> player1builder;
    /**
     * player strategy 1.
     */
    private Optional<VierGewinntStrategy> player1Strategy;
    /**
     * player builder 2.
     */
    private Optional<VierGewinntPlayerBuilder> player2builder;
    /**
     * player strategy 2.
     */
    private Optional<VierGewinntStrategy> player2Strategy;

    /**
     * Constructor.
     */
    public VierGewinntGameBuilderImpl() {
        this.player1builder = Optional.empty();
        this.player1Strategy = Optional.empty();
        this.player2builder = Optional.empty();
        this.player2Strategy = Optional.empty();
        this.newObserverFactoryProvider = new DefaultObserverFactoryProvider();
        this.maxComputationTimePerMove = 5;
    }

    @Override
    public GameBuilder changeMaximumComputationTimePerMove(final int newMaxComputationTimePerMove) {
        this.maxComputationTimePerMove = newMaxComputationTimePerMove;
        return this;
    }

    @Override
    public VierGewinntPlayerBuilder createPlayerBuilder() {
        return new VierGewinntPlayerBuilderImpl();
    }

    @Override
    public VierGewinntGameBuilder addPlayerBuilder(final VierGewinntPlayerBuilder playerBuilder,
            final VierGewinntStrategy strategy) throws GameException {
        Objects.requireNonNull(playerBuilder);
        if (this.player1builder.isEmpty()) {
            this.player1builder = Optional.of(playerBuilder);
            this.player1Strategy = Optional.of(Objects.requireNonNull(strategy, "firstPlayerStrategy"));
        } else if (this.player2builder.isEmpty()) {
            this.player2builder = Optional.of(playerBuilder);
            this.player2Strategy = Optional.of(Objects.requireNonNull(strategy, "secondPlayerStrategy"));
        } else {
            throw new GameException(String.format("More than two players are now allowed."));
        }
        return this;
    }

    @Override
    public VierGewinntGameBuilder changeColumnSize(final int newColumnSize) {
        this.columnSize = newColumnSize;
        return this;
    }

    @Override
    public VierGewinntGameBuilder changeRowSize(final int newRowSize) {
        this.rowSize = newRowSize;
        return this;
    }

    @Override
    public VierGewinntGameBuilder changeObserverFactoryProvider(final ObserverFactoryProvider provider) {
        this.newObserverFactoryProvider = provider;
        return this;
    }

    @Override
    public VierGewinntGame build(final int id)
            throws GameException, IllegalArgumentException, InterruptedException {
        if (!this.player1builder.isPresent() || !this.player2builder.isPresent()) {
            throw new GameException("A game needs two players!");
        }

        final VierGewinntState initialState =
                new VierGewinntStateImpl(new VierGewinntBoardImpl(rowSize, columnSize), this.player1builder.get(),
                        this.player2builder.get());

        final Map<String, VierGewinntStrategy> strategies = new LinkedHashMap<>();
        strategies.put(initialState.getYellowPlayer().getName(), this.player1Strategy.orElseThrow());
        strategies.put(initialState.getRedPlayer().getName(), this.player2Strategy.orElseThrow());
        return new VierGewinntGameImpl(id, initialState, strategies, this.maxComputationTimePerMove,
                AbstractMove.class::isInstance, this.newObserverFactoryProvider);
    }

}
