package de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import de.fhdw.gaming.core.domain.GameException;
import de.fhdw.gaming.core.domain.PlayerState;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntBoard;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntDirection;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntField;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntFieldState;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntPlayer;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntPlayerBuilder;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.VierGewinntState;
import de.fhdw.gaming.ipspiel21.viergewinnt.core.domain.moves.VierGewinntMove;

/**
 * The Implementation of a VierGewinntState.
 *
 * @author Robby Rabbitman
 *
 */
public class VierGewinntStateImpl implements VierGewinntState {
    /**
     * 4 in row.
     *
     */
    private static final Integer WIN_CONDITION = 4;

    /**
     * The board.
     */
    private final VierGewinntBoard board;
    /**
     * The player using yellow.
     */
    private final VierGewinntPlayer yellowPlayer;
    /**
     * The player using red.
     */
    private final VierGewinntPlayer redPlayer;
    /**
     * The current player.
     */
    private VierGewinntPlayer currentPlayer;
    /**
     * The states of the players.
     */
    private final Map<String, PlayerState> playerStates;
    /**
     * The outcomes of the players.
     */
    private final Map<String, Double> playerOutcomes;

    /**
     * .
     *
     */
    private Boolean gameOver;

    /**
     * The Constructor.
     *
     * @param board
     * @param yellowPlayer
     * @param redPlayer
     * @param currentPlayer
     * @param playerStates
     * @param playerOutcomes
     */
    public VierGewinntStateImpl(final VierGewinntBoard board, final VierGewinntPlayer yellowPlayer,
            final VierGewinntPlayer redPlayer, final VierGewinntPlayer currentPlayer,
            final Map<String, PlayerState> playerStates, final Map<String, Double> playerOutcomes) {
        super();
        this.board = board;
        this.gameOver = false;
        this.yellowPlayer = yellowPlayer;
        this.redPlayer = redPlayer;
        this.currentPlayer = currentPlayer;
        this.playerStates = playerStates;
        this.playerOutcomes = playerOutcomes;
    }

    /**
     * Creates an VierGewinnt state by copying an existing one.
     *
     * @param source The state to copy.
     */
    VierGewinntStateImpl(final VierGewinntStateImpl source) {
        this.board = source.getBoard().deepCopy();
        this.yellowPlayer = source.yellowPlayer.deepCopy(this);
        this.redPlayer = source.redPlayer.deepCopy(this);
        this.currentPlayer = source.isYellowPlayerCurrent() ? this.yellowPlayer : this.redPlayer;
        this.playerOutcomes = new LinkedHashMap<>(source.playerOutcomes);
        this.playerStates = new LinkedHashMap<>();
        this.playerStates.put(this.redPlayer.getName(), source.playerStates.get(this.redPlayer.getName()));
        this.playerStates.put(this.yellowPlayer.getName(), source.playerStates.get(this.yellowPlayer.getName()));
        this.gameOver = source.gameOver;
    }

    /**
     * .
     *
     * @param board
     * @param player1builder
     * @param player2builder
     * @throws GameException
     */
    public VierGewinntStateImpl(final VierGewinntBoardImpl board, final VierGewinntPlayerBuilder player1builder,
            final VierGewinntPlayerBuilder player2builder) throws GameException {
        this.board = Objects.requireNonNull(board, "board");
        this.yellowPlayer = Objects.requireNonNull(player1builder, "crossesPlayerBuilder").build(this);
        this.redPlayer = Objects.requireNonNull(player2builder, "noughtsPlayerBuilder").build(this);
        this.currentPlayer = this.yellowPlayer;
        this.playerOutcomes = new LinkedHashMap<>();

        this.gameOver = false;
        this.playerStates = new LinkedHashMap<>();
        this.playerStates.put(this.yellowPlayer.getName(), PlayerState.PLAYING);
        this.playerStates.put(this.redPlayer.getName(), PlayerState.PLAYING);

        if (!this.yellowPlayer.isUsingYellow()) {
            throw new IllegalArgumentException(String.format("Player %s does not use yellow.", this.yellowPlayer));
        }
        if (this.redPlayer.isUsingYellow()) {
            throw new IllegalArgumentException(String.format("Player %s does not use red.", this.redPlayer));
        }
        if (this.yellowPlayer.getName().equals(this.redPlayer.getName())) {
            throw new IllegalArgumentException(
                    String.format("Both players have the same name '%s'.", this.yellowPlayer.getName()));
        }
    }

    @Override
    public Map<String, VierGewinntPlayer> getPlayers() {
        return Map.ofEntries(Map.entry(this.getYellowPlayer().getName(), this.getYellowPlayer()),
                Map.entry(this.getRedPlayer().getName(), this.getRedPlayer()));
    }

    @Override
    public PlayerState getPlayerState(final String playerName) throws IllegalArgumentException {
        final PlayerState playerState = this.playerStates.get(playerName);
        if (playerState != null) {
            return playerState;
        } else {
            throw new IllegalArgumentException(String.format("Unknown player %s.", playerName));
        }
    }

    @Override
    public void setPlayerState(final String playerName, final PlayerState newState)
            throws IllegalArgumentException {
        if (this.playerStates.containsKey(playerName)) {
            this.playerStates.put(playerName, newState);
            if (newState.equals(PlayerState.PLAYING)) {
                this.playerOutcomes.remove(playerName);
            }
        } else {
            throw new IllegalArgumentException(String.format("Unknown player %s.", playerName));
        }
    }

    @Override
    public void setPlayerOutcome(final String playerName, final double newOutcome)
            throws IllegalArgumentException {
        if (this.getPlayerState(playerName).equals(PlayerState.PLAYING)) {
            throw new IllegalArgumentException(String.format("Cannot set outcome for player %s.", playerName));
        } else {
            this.playerOutcomes.put(playerName, newOutcome);
        }
    }

    @Override
    public Set<VierGewinntPlayer> computeNextPlayers() {
        if (this.gameOver) {
            return Set.of();
        } else {
            return Set.of(this.currentPlayer);
        }
    }

    @Override
    public void nextTurn() {
        this.currentPlayer = this.getOtherPlayer();
    }

    @Override
    public VierGewinntState deepCopy() {
        return new VierGewinntStateImpl(this);
    }

    @Override
    public VierGewinntBoard getBoard() {
        return this.board;
    }

    @Override
    public VierGewinntPlayer getYellowPlayer() {
        return this.yellowPlayer;
    }

    @Override
    public VierGewinntPlayer getRedPlayer() {
        return this.redPlayer;
    }

    @Override
    public VierGewinntPlayer getCurrentPlayer() {
        return this.currentPlayer;
    }

    @Override
    public void moveCompleted(final VierGewinntMove move) {
        // check if draw
        final boolean draw = this.board.getFieldsBeing(VierGewinntFieldState.EMPTY).size() == 0;
        if (draw) {
            this.gameOver = true;
            this.setPlayerState(this.currentPlayer.getName(), PlayerState.DRAW);
            this.setPlayerState(this.getOtherPlayer().getName(), PlayerState.DRAW);
        } else {
            // check if current player won
            final boolean currentPlayerWon = this
                    .getRows(this.getBoard().getFieldAt(move.getPosition()), VierGewinntStateImpl.WIN_CONDITION)
                    .stream().map(row -> this.checkGameOver(row, this.currentPlayer.getPlayerMark(),
                            VierGewinntStateImpl.WIN_CONDITION))
                    .anyMatch(gameover -> gameover.equals(true));
            if (currentPlayerWon) {
                this.gameOver = true;
                this.setPlayerState(this.currentPlayer.getName(), PlayerState.WON);
                this.setPlayerState(this.getOtherPlayer().getName(), PlayerState.LOST);
            }
        }
    }

    /**
     * Checks if a game is over or not.
     *
     * @param row
     * @param state
     * @param length
     * @return
     */
    private Boolean checkGameOver(final List<VierGewinntField> row, final VierGewinntFieldState state,
            final Integer length) {
        boolean won = false;
        for (int i = 0; i <= row.size() - length; i++) {
            won = true;
            for (int j = 0; j < length; j++) {
                won = won && row.get(j + i).getState().name().equals(state.name());
            }
            if (won) {
                return won;
            }
        }
        return won;
    }

    /**
     * Return a nested list of the rows of a VierGewinntGame.
     *
     * @param field
     * @param radius
     * @return
     */
    private List<List<VierGewinntField>> getRows(final VierGewinntField field, final Integer radius) {
        final List<List<VierGewinntField>> rows = new ArrayList<>();
        rows.add(this.getRowsHelperConcat(this.getRowsHelper(field, radius, VierGewinntDirection.NORTHWEST), field,
                this.getRowsHelper(field, radius, VierGewinntDirection.SOUTHEAST)));
        rows.add(this.getRowsHelperConcat(this.getRowsHelper(field, radius, VierGewinntDirection.NORTHEAST), field,
                this.getRowsHelper(field, radius, VierGewinntDirection.SOUTHWEST)));
        rows.add(this.getRowsHelperConcat(this.getRowsHelper(field, radius, VierGewinntDirection.NORTH), field,
                this.getRowsHelper(field, radius, VierGewinntDirection.SOUTH)));
        rows.add(this.getRowsHelperConcat(this.getRowsHelper(field, radius, VierGewinntDirection.WEST), field,
                this.getRowsHelper(field, radius, VierGewinntDirection.EAST)));

        return rows;
    }

    /**
     * Concat the return of the getRowsHelperfunction to a nested List.
     *
     * @param list1
     * @param field
     * @param list2
     * @return
     */
    private List<VierGewinntField> getRowsHelperConcat(final List<VierGewinntField> list1,
            final VierGewinntField field, final List<VierGewinntField> list2) {
        final List<VierGewinntField> concat = new ArrayList<>();
        concat.addAll(list1);
        Collections.reverse(concat);
        concat.add(field);
        concat.addAll(list2);
        return concat;

    }

    /**
     * Helper to build the nested rows for a VierGewinntGame.
     *
     * @param field
     * @param radius
     * @param direction
     * @return
     */
    private List<VierGewinntField> getRowsHelper(final VierGewinntField field, final Integer radius,
            final VierGewinntDirection direction) {
        final List<VierGewinntField> out = new ArrayList<>();
        VierGewinntField currentField = field;
        for (int i = 0; i < radius - 1; i++) {
            if (currentField.hasNeighbour(direction)) {
                currentField = currentField.getNeighbour(direction);
                out.add(currentField);
            } else {
                return out;
            }
        }
        return out;
    }

    /**
     * Returns the currently inactive player.
     */
    private VierGewinntPlayer getOtherPlayer() {
        if (this.isYellowPlayerCurrent()) {
            return this.redPlayer;
        } else {
            return this.yellowPlayer;
        }
    }

    /**
     * Return if the Yellow Player is current.
     *
     * @return
     */
    private boolean isYellowPlayerCurrent() {
        return this.currentPlayer.equals(this.yellowPlayer);
    }

}
